@@ -101,13 +101,22 @@ def resolve(type_: ast.Type):
     return t
 
 
+def maybe_autocommit():
+    if context.state.autocommit:
+        get_db().commit()
 
-def db_query(sql_code, subqueries=None):
+
+def db_query(sql_code, subqueries=None, *, modifies=True):
     try:
-        return get_db().query(sql_code, subqueries)
+        res = get_db().query(sql_code, subqueries)
     except exc.DatabaseQueryError as e:
         raise Signal.make(T.DbQueryError, None, e.args[0]) from e
 
+    if modifies:
+        maybe_autocommit()
+
+    return res
+
 def drop_table(table_type):
     name = table_type.options['name']
     code = sql.compile_drop_table(name)
@@ -232,10 +241,11 @@ def _execute(table_def: ast.TableDef):
             t = t({**t.elems, **elems_to_add}, **cur_type.options)
 
         # Auto-add id only if it exists already and not defined by user
-        if 'id' in cur_type.elems: #and 'id' not in t.elems:
-            # t = t(dict(id=T.t_id, **t.elems), pk=[['id']])
-            assert cur_type.elems['id'] <= T.primitive, cur_type.elems['id']
-            t.elems['id'] = T.t_id
+        if not table_def.bare:
+            if 'id' in cur_type.elems: #and 'id' not in t.elems:
+                # t = t(dict(id=T.t_id, **t.elems), pk=[['id']])
+                assert cur_type.elems['id'] <= T.primitive, cur_type.elems['id']
+                t.elems['id'] = T.t_id
 
         for e_name, e1_type in t.elems.items():
 
@@ -251,9 +261,10 @@ def _execute(table_def: ast.TableDef):
     else:
         # Auto-add id by default
         elems = dict(t.elems)
-        if 'id' not in elems:
-            elems = {'id': T.t_id, **elems}
-        t = t(elems, pk=[['id']])
+        if not table_def.bare:
+            if 'id' not in elems:
+                elems = {'id': T.t_id, **elems}
+            t = t(elems, pk=[['id']])
         inst = objects.new_table(t, db_name)
 
     ensure_namespace(table_def.name)
@@ -344,6 +355,19 @@ def _execute(f: ast.For):
         with use_scope({f.var: objects.from_python(i)}):
             execute(f.do)
 
+@method
+def _execute(t: ast.Transaction):
+    db_query(sql.BeginTransaction())
+    with context(state=context.state.set_autocommit(False)):
+        try:
+            res = execute(t.do)
+        except Signal as e:
+            get_db().rollback()
+            raise
+
+    get_db().commit()
+    return res
+
 @method
 def _execute(t: ast.Try):
     try:
@@ -367,7 +391,7 @@ def find_module(module_name):
         if module_path.exists():
             return module_path
 
-    raise Signal.make(T.ImportError, r, "Cannot find module")
+    raise Signal.make(T.ImportError, module_name, "Cannot find module")
 
 
 def import_module(state, r):
@@ -408,7 +432,8 @@ def _execute(t: ast.Throw):
     e = evaluate(t.value)
     if isinstance(e, ast.Ast):
         raise exc.InsufficientAccessLevel()
-    assert isinstance(e, Exception), e
+    if not isinstance(e, Exception):
+        raise Signal.make(T.TypeError, t, f"Can only throw an exception, not {e.type}")
     raise e
 
 def execute(stmt):
@@ -684,13 +709,31 @@ def apply_database_rw(o: ast.One):
     return objects.RowInstance(rowtype, d)
 
 
+def _delete_from_table(table, conds):
+    assert isinstance(table.code, sql.TableName)
+    with use_scope({n:objects.projected(c) for n, c in table.all_attrs().items()}):
+        compiled_conds = cast_to_instance(conds) 
+
+    code = sql.Delete(table.code, [c.code for c in compiled_conds])
+    db_query(code, table.subqueries)
+
 @method
 def apply_database_rw(d: ast.Delete):
     catch_access(AccessLevels.WRITE_DB)
-    # TODO Optimize: Delete on condition, not id, when possible
+
+    if isinstance(d.table, ast.Selection):
+        table = evaluate(d.table.table)
+        if isinstance(table.code, sql.TableName):
+            _delete_from_table(table, d.table.conds + d.conds)
+            return objects.null
+
+    table = evaluate(d.table)
+    if isinstance(table.code, sql.TableName):
+        _delete_from_table(table, d.conds)
+
 
     cond_table = ast.Selection(d.table, d.conds).set_text_ref(d.text_ref)
-    table = evaluate( cond_table)
+    table = evaluate(cond_table)
 
     if not table.type <= T.table:
         raise Signal.make(T.TypeError, d.table, f"Expected a table. Got: {table.type}")
@@ -714,8 +757,22 @@ def apply_database_rw(d: ast.Delete):
 def apply_database_rw(u: ast.Update):
     catch_access(AccessLevels.WRITE_DB)
 
+    if isinstance(u.table, ast.Selection):
+        table = evaluate(u.table.table)
+        if isinstance(table.code, sql.TableName):
+
+            with use_scope({n:objects.projected(c) for n, c in table.all_attrs().items()}):
+                conds = cast_to_instance(u.table.conds)
+                proj = {f.name:evaluate( f.value) for f in u.fields}
+
+            # breakpoint()
+            code = sql.Update(table.code, {sql.Name(v.type, k): v.code for k, v in proj.items()}, [c.code for c in conds])
+            db_query(code, table.subqueries)
+
+            return objects.null
+
     # TODO Optimize: Update on condition, not id, when possible
-    table = evaluate( u.table)
+    table = evaluate(u.table)
 
     if not table.type <= T.table:
         raise Signal.make(T.TypeError, u.table, f"Expected a table. Got: {table.type}")
@@ -836,9 +893,11 @@ def _new_row(new_ast, table, matched):
         raise Signal.make(T.TypeError, new_ast, f"'new' expects a persistent table. Instead got a table expression.")
 
     if get_db().target == sql.bigquery:
-        rowid = db_query(sql.FuncCall(T.string, 'GENERATE_UUID', []))
+        rowid = db_query(sql.FuncCall(T.string, 'GENERATE_UUID', []), modifies=False)
         keys += ['id']
         values += [sql.make_value(rowid)]
+    elif get_db().target == sql.snowflake:
+        rowid = None
 
     try:
         table_name = table.options['name']
@@ -848,9 +907,10 @@ def _new_row(new_ast, table, matched):
     q = sql.InsertConsts(table_name, keys, [values])
     db_query(q)
 
-    if get_db().target != sql.bigquery:
-        rowid = db_query(sql.LastRowId())
+    if get_db().target in (sql.bigquery, sql.snowflake, sql.presto):
+        return objects.null
 
+    rowid = db_query(sql.LastRowId(), modifies=False)
     d = SafeDict({'id': objects.pyvalue_inst(rowid)})
     d.update({p.name:v for p, v in matched})
     return objects.RowInstance(T.row[table], d)
@@ -979,7 +1039,7 @@ def localize(inst: objects.Instance):
     if inst.code is sql.null:
         return None
 
-    return db_query(inst.code, inst.subqueries)
+    return db_query(inst.code, inst.subqueries, modifies=False)
 
 @method
 def localize(inst: objects.ValueInstance):
