@@ -18,7 +18,7 @@ from . import pql_ast as ast
 from . import sql
 from .interp_common import pyvalue_inst, assert_type, cast_to_python_string, cast_to_python_int, cast_to_python
 from .state import get_var, get_db, use_scope, unique_name, get_db, require_access, AccessLevels, set_var
-from .evaluate import evaluate, db_query, TableConstructor, new_table_from_expr, new_table_from_rows
+from .evaluate import evaluate, db_query, maybe_autocommit, TableConstructor, new_table_from_expr, new_table_from_rows
 from .pql_types import T, Type, Id
 from .types_impl import join_names
 from .casts import cast
@@ -215,6 +215,18 @@ def pql_set_setting(name: T.string, value: T.any):
     setattr(Display, name, value)
     return objects.null
 
+def pql_set_active_dataset(name: T.string):
+    name = cast_to_python_string(name)
+    db = get_db()
+    db.set_active_dataset(name)
+    return objects.null
+
+def pql_get_qualified_name(name: T.string):
+    name = cast_to_python_string(name)
+    db = get_db()
+    qualified_name = sql.quote_id(db.qualified_name(Id(name)))
+    return pyvalue_inst(qualified_name)
+
 
 def pql_debug():
     """Breaks the execution of the interpreter, and enters into a debug
@@ -857,6 +869,7 @@ def import_pandas(dfs):
 
         tbl = new_table_from_rows(name, cols, rows, temporary=True)
         set_var(name, tbl)
+        maybe_autocommit()
         yield tbl
 
 def pql_import_json(table_name: T.string, uri: T.string):
@@ -900,7 +913,7 @@ def pql_import_csv(table: T.table, filename: T.string, header: T.bool = ast.Cons
     header = cast_to_python(header)
     msg = f"Importing CSV file: '{filename}'"
 
-    ROWS_PER_QUERY = 1024
+    max_rows_per_query = get_db().max_rows_per_query
 
     cons = TableConstructor.make(table.type)
     keys = []
@@ -929,18 +942,21 @@ def pql_import_csv(table: T.table, filename: T.string, header: T.bool = ast.Cons
                 values = ["'%s'" % (v.replace("'", "''")) for v in row]
                 rows.append(values)
 
-                if (i+1) % ROWS_PER_QUERY == 0:
+                if (i+1) % max_rows_per_query == 0:
                     insert_values()
                     rows = []
 
         if keys and rows:
             insert_values()
 
+        maybe_autocommit()
+
         return table
     except FileNotFoundError as e:
         raise Signal.make(T.FileError, None, str(e))
 
 
+
 def _rest_func_endpoint(func):
     from starlette.responses import JSONResponse
     async def callback(request):
@@ -1024,7 +1040,7 @@ def pql_serve_rest(endpoints: T.struct, port: T.int = pyvalue_inst(8080)):
     return objects.null
 
 
-def pql_table_add_index(table, column_name: T.string, unique: T.bool = ast.false):
+def pql_table_add_index(table, column_name: T.string | T.list[T.string], unique: T.bool = ast.false):
     """Add an index to the table, to optimize filtering operations.
 
     A method of the `table` type.
@@ -1040,6 +1056,10 @@ def pql_table_add_index(table, column_name: T.string, unique: T.bool = ast.false
         >> table x = [1,2,3]{item}
         >> x.add_index("item")
     """
+    if get_db().target in (sql.snowflake, sql.redshift, sql.bigquery):
+        # TODO 
+        return objects.null
+
     try:
         table_name = table.type.options['name']
     except KeyError:
@@ -1047,11 +1067,13 @@ def pql_table_add_index(table, column_name: T.string, unique: T.bool = ast.false
 
     unique = cast_to_python(unique)
     column_name = cast_to_python(column_name)
+    if isinstance(column_name, str):
+        column_name = [column_name]
     
-    index_name = unique_name(f'index_{table_name.repr_name}_{column_name}')
+    index_name = unique_name(f'index_{table_name.repr_name}_{"_".join(column_name)}')
 
     code = sql.AddIndex(Id(index_name), table_name, column_name, unique=unique)
-    db_query(code)
+    db_query(code, modifies=True)
 
     return objects.null
 
@@ -1105,6 +1127,8 @@ internal_funcs = create_internal_funcs({
     'force_eval': pql_force_eval,
     'fmt': pql_fmt,
     'set_setting': pql_set_setting,
+    'set_active_dataset': pql_set_active_dataset,
+    'get_qualified_name': pql_get_qualified_name,
 })
 
 _joins = {
