@@ -5,7 +5,7 @@ from preql.utils import dataclass, X, listgen, safezip
 from . import pql_types
 from .pql_types import T, Type, dp_type, Id
 from .types_impl import join_names, flatten_type
-from .state import get_db
+from .state import get_db, unique_name
 from .exceptions import Signal
 
 duck = 'duck'
@@ -13,6 +13,10 @@ sqlite = 'sqlite'
 postgres = 'postgres'
 bigquery = 'bigquery'
 mysql = 'mysql'
+snowflake = 'snowflake'
+redshift = 'redshift'
+presto = 'presto'
+oracle = 'oracle'
 
 class QueryBuilder:
     def __init__(self, is_root=True, start_count=0):
@@ -64,7 +68,7 @@ class SqlTree(Sql):
     def finalize_with_subqueries(self, qb, subqueries):
         if subqueries:
             subqs = [q.compile_wrap(qb).finalize(qb) for (name, q) in subqueries.items()]
-            sql_code = ['WITH RECURSIVE '] if qb.target in (postgres, mysql) else ['WITH ']
+            sql_code = ['WITH RECURSIVE '] if qb.target in (postgres, mysql, redshift) else ['WITH ']
             sql_code += join_comma([q, '\n    '] for q in subqs)
         else:
             sql_code = []
@@ -85,6 +89,8 @@ class CompiledSQL(Sql):
         assert qb.is_root
         if wrapped.type <= T.primitive and not wrapped.code[0].lower().startswith('select '):
             code = ['SELECT '] + wrapped.code
+            if get_db().target == oracle:
+                code += [' FROM dual']
         else:
             code = wrapped.code
         return ''.join(code)
@@ -240,7 +246,7 @@ class JsonLength(Scalar):
         code = self.expr.compile_wrap(qb).code
         if qb.target == sqlite:
             return [f'(length('] + code + [') - length(replace('] + code + [f', "{_ARRAY_SEP}", ""))) / length("{_ARRAY_SEP}") + 1']
-        elif qb.target == postgres:
+        elif qb.target in (postgres, redshift):
             return [f'array_length('] + code + [', 1)']
         elif qb.target == bigquery:
             return [f'array_length('] + code + [')']
@@ -299,7 +305,7 @@ class MakeArray(SqlTree):
         field = self.field.compile_wrap(qb).code
         if qb.target in (sqlite, duck):
             return ['group_concat('] + field + [f', "{_ARRAY_SEP}")']
-        elif qb.target == postgres:
+        elif qb.target in (postgres, redshift):
             return ['array_agg('] + field + [')']
         elif qb.target == mysql:
             return ['json_arrayagg('] + field + [')']
@@ -488,18 +494,29 @@ class ColumnAlias(SqlTree):
 class SqlStatement(SqlTree):
     type = T.nulltype
 
+@dataclass
+class BeginTransaction(SqlStatement):
+    def _compile(self, qb):
+        if qb.target == mysql:
+            return ["START TRANSACTION"]
+        return ["BEGIN TRANSACTION"]
+
 @dataclass
 class AddIndex(SqlStatement):
     index_name: Id
     table_name: Id
-    column: str
+    columns: List[str]
     unique: bool
 
     def _compile(self, qb):
+        if qb.target == redshift:
+            # TODO issue some warning? error?
+            return ['SELECT NULL']
+
         stmt = f"CREATE {'UNIQUE' if self.unique else ''} INDEX "
-        if qb.target != mysql:
+        if qb.target not in (mysql, oracle):
             stmt += "IF NOT EXISTS "
-        return [ stmt + f"{quote_id(self.index_name)} ON {quote_id(self.table_name)}({self.column})"]
+        return [ stmt + f"{quote_id(self.index_name)} ON {quote_id(self.table_name)}({', '.join(self.columns)})"]
 
 @dataclass
 class Insert(SqlStatement):
@@ -512,7 +529,7 @@ class Insert(SqlStatement):
         return [f'INSERT INTO {quote_id(self.table_name)}({", ".join(columns)}) '] + self.query.compile(qb).code
 
     def finalize_with_subqueries(self, qb, subqueries):
-        if qb.target in (mysql, bigquery):
+        if qb.target in (mysql, bigquery, oracle):
             columns = [quote_name(c) for c in self.columns]
             sql_code = f'INSERT INTO {quote_id(self.table_name)}({", ".join(columns)}) '
             sql_code += self.query.finalize_with_subqueries(qb, subqueries)
@@ -555,17 +572,18 @@ class InsertConsts2(SqlStatement):
     def _compile(self, qb):
         assert self.tuples, self
 
-        values = join_comma(
-            parens(join_comma([t] for t in tpl))
-            for tpl in self.tuples
-        )
+
+        if get_db().target == oracle:
+            values_q = join_sep([['SELECT '] + join_comma([t] for t in tpl) + [' FROM dual'] for tpl in self.tuples], ' UNION ALL ')
+        else:
+            values_q = ['VALUES '] + join_comma(
+                parens(join_comma([t] for t in tpl))
+                for tpl in self.tuples
+            )
 
         cols = [quote_name(c) for c in self.cols]
-        q = ['INSERT INTO', quote_id(self.table),
-             "(", ', '.join(cols), ")",
-             "VALUES ",
-        ]
-        return [' '.join(q)] + values #+ ';'
+        q = ['INSERT INTO ' , quote_id(self.table), "(", ', '.join(cols), ")", ] + values_q
+        return q
 
 
 @dataclass
@@ -587,7 +605,8 @@ class SelectValue(Atom, TableOperation):
 
     def _compile(self, qb):
         value = self.value.compile_wrap(qb)
-        return [f'SELECT '] + value.code + [' AS '] + value
+        code = [f'SELECT '] + value.code + [' AS '] + value
+        return code
 
     type = property(X.value.type)
 
@@ -606,6 +625,10 @@ class Values(Table):
         if qb.target == mysql:
             def row_func(x):
                 return ['ROW('] + x + [')']
+        elif qb.target in (snowflake, redshift):
+            return join_sep([['SELECT '] + v.code for v in values], ' UNION ALL ')
+        elif qb.target in (oracle, ):
+            return join_sep([['SELECT '] + v.code + [' FROM dual'] for v in values], ' UNION ALL ')
         else:
             row_func = parens
 
@@ -638,6 +661,7 @@ class ValuesTuples(Table):
         if not self.values:  # SQL doesn't support empty values
             return ['SELECT '] + join_comma(['NULL']*len(self.type.elems)) +  ['LIMIT 0']
         values = [v.compile_wrap(qb) for v in self.values]
+
         return ['VALUES '] + join_comma(v.code for v in values)
 
 
@@ -725,6 +749,8 @@ class Select(TableOperation):
         select_sql = join_comma(f.code for f in fields_sql)
 
         sql = ['SELECT '] + select_sql + [' FROM '] + self.table.compile_wrap(qb).code
+        if qb.target == bigquery:
+            sql += [' ', qb.unique_name()]
 
         if self.conds:
             sql += [' WHERE '] + join_sep([c.compile_wrap(qb).code for c in self.conds], ' AND ')
@@ -736,23 +762,33 @@ class Select(TableOperation):
         if self.order:
             sql += [' ORDER BY '] + join_comma(o.compile_wrap(qb).code for o in self.order)
 
-        if self.limit is not None:
-            sql += [' LIMIT ', str(self.limit)]
-        elif self.offset is not None:
-            if qb.target == sqlite:
-                sql += [' LIMIT -1']  # Sqlite only (and only old versions of it)
-            elif qb.target == mysql:
-                # MySQL requires a specific limit, always!
-                # See: https://stackoverflow.com/questions/255517/mysql-offset-infinite-rows
-                sql += [' LIMIT 18446744073709551615']
-            elif qb.target == bigquery:
-                # BigQuery requires a specific limit, always!
-                sql += [' LIMIT 9223372036854775807']
-
-
-        if self.offset is not None:
+        if self.offset is not None and get_db().offset_before_limit:
             sql += [' OFFSET ', str(self.offset)]
 
+        if qb.target == oracle:
+            if self.offset is not None:
+                sql += [' OFFSET ', str(self.offset), ' ROWS']
+
+            if self.limit is not None:
+                sql += [' FETCH NEXT ', str(self.limit), ' ROWS ONLY']
+        
+        else:
+            if self.limit is not None:
+                sql += [' LIMIT ', str(self.limit)]
+            elif self.offset is not None:
+                if qb.target == sqlite:
+                    sql += [' LIMIT -1']  # Sqlite only (and only old versions of it)
+                elif qb.target == mysql:
+                    # MySQL requires a specific limit, always!
+                    # See: https://stackoverflow.com/questions/255517/mysql-offset-infinite-rows
+                    sql += [' LIMIT 18446744073709551615']
+                elif qb.target == bigquery:
+                    # BigQuery requires a specific limit, always!
+                    sql += [' LIMIT 9223372036854775807']
+
+            if self.offset is not None and not get_db().offset_before_limit:
+                sql += [' OFFSET ', str(self.offset)]
+
         return sql
 
 
@@ -879,7 +915,7 @@ class StringSlice(SqlTree):
             params = string + [', '] + start
             if length:
                 params += [', '] + length
-        elif qb.target in (postgres, mysql):
+        elif qb.target in (postgres, redshift, mysql):
             f = 'substring'
             params = string + [' from '] + start
             if length:
@@ -896,6 +932,8 @@ def _repr(_t: T.number, x):
 
 @dp_type
 def _repr(_t: T.bool, x):
+    if get_db().target == oracle:
+        return ['1=0', '1=1'][x]
     return ['false', 'true'][x]
 
 @dp_type
@@ -965,6 +1003,12 @@ class P2S_Sqlite(Types_PqlToSql):
 class P2S_Postgres(Types_PqlToSql):
     datetime = "timestamp with time zone"
 
+class P2S_Snowflake(Types_PqlToSql):
+    pass
+
+class P2S_Oracle(Types_PqlToSql):
+    pass
+
 _pql_to_sql_by_target = {
     bigquery: P2S_BigQuery,
     mysql: P2S_MySql,
@@ -972,6 +1016,10 @@ _pql_to_sql_by_target = {
     sqlite: P2S_Sqlite,
     duck: P2S_Sqlite,
     postgres: P2S_Postgres,
+    redshift: P2S_Postgres,
+    snowflake: P2S_Snowflake,
+    presto: P2S_MySql,
+    oracle: P2S_Oracle,
 }
 
 
@@ -1024,7 +1072,7 @@ def compile_type_def(table_name, table) -> Sql:
     pks = []
     columns = []
 
-    pks = {join_names(pk) for pk in table.options['pk']}
+    pks = {join_names(pk) for pk in table.options.get('pk', [])}
     for name, c in flatten_type(table):
         if name in pks and c <= T.t_id:
             type_decl = db.id_type_decl
@@ -1051,6 +1099,8 @@ def compile_type_def(table_name, table) -> Sql:
     # Consistent among SQL databases
     if db.target == 'bigquery':
         command = ("CREATE TABLE" if table.options.get('temporary', False) else "CREATE TABLE IF NOT EXISTS")
+    elif db.target == 'oracle':
+        command = "CREATE TEMPORARY TABLE" if table.options.get('temporary', False) else "CREATE TABLE"
     else:
         command = "CREATE TEMPORARY TABLE" if table.options.get('temporary', False) else "CREATE TABLE IF NOT EXISTS"
 
@@ -1109,11 +1159,14 @@ def arith(res_type, op, args):
             # In MySQL division returns a float. All others return int
             arg_codes[0] = Cast(T.float, arg_codes[0])
     elif op == '/~':
-        if target == mysql:
+        if target in (mysql, oracle):
             op = 'DIV'
         elif target == bigquery:
             # safe div?
             return FuncCall(res_type, 'div', arg_codes)
+        elif target == snowflake:
+            res = BinOp(res_type, '/', arg_codes)
+            return FuncCall(T.int, 'floor', [res])
         else:
             op = '/'
     elif op == '%':
