@@ -9,7 +9,7 @@ from .utils import classify, dataclass
 from .loggers import sql_log
 from .context import context
 
-from .core.sql import Sql, QueryBuilder, sqlite, postgres, mysql, duck, bigquery, quote_id
+from .core.sql import Sql, QueryBuilder, sqlite, postgres, mysql, duck, bigquery, quote_id, snowflake, redshift, oracle, presto
 from .core.sql_import_result import sql_result_to_python, type_from_sql
 from .core.pql_types import T, Type, Object, Id
 from .core.exceptions import DatabaseQueryError, Signal
@@ -35,6 +35,8 @@ class SqlInterface:
     supports_foreign_key = True
     requires_subquery_name = False
     id_type_decl = 'INTEGER'
+    max_rows_per_query = 1024
+    offset_before_limit = False
 
 
     def __init__(self, print_sql=False):
@@ -80,6 +82,9 @@ class SqlInterface:
     def quote_name(self, name):
         return f'"{name}"'
 
+    def set_active_dataset(self, name):
+        raise Signal.make(T.DbError, None, "set_active_dataset() is not implemented for %s" % type(self).__name__)
+
 
 
 # from multiprocessing import Queue
@@ -210,6 +215,58 @@ class SqlInterfaceCursor(SqlInterface):
         assert n == 1
 
 
+class OracleInterface(SqlInterfaceCursor):
+    target = oracle
+
+    id_type_decl = "NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY"
+
+    def __init__(self, host, port, database, user, password, print_sql=False):
+        self._print_sql = print_sql
+        self.args = dict(dsn="%s/%s" % (host, database), user=user, password=password)
+
+        super().__init__(print_sql)
+
+    def _create_connection(self):
+        import cx_Oracle
+        try:
+            # return cx_Oracle.connect(user="oracle", password="TzR2n6e4i7ULgSzN", dsn="127.0.0.1/ORACLE")
+            return cx_Oracle.connect(**self.args)
+        except Exception as e:
+            raise ConnectError(*e.args) from e
+
+    def list_tables(self):
+        sql_code = "SELECT table_name FROM all_tables WHERE tablespace_name = 'USERS' ORDER BY table_name"
+        names = self._execute_sql(T.list[T.string], sql_code)
+        return list(map(Id, names))
+
+    def import_table_type(self, name, columns_whitelist=None):
+
+        columns_t = T.table(dict(
+            name=T.string,
+            type=T.string,
+            nullable=T.string,
+            default=T.string,
+        ))
+        columns_q = "SELECT column_name, data_type, nullable, data_default FROM USER_TAB_COLUMNS WHERE table_name = '%s'" % name.name.upper()
+        sql_columns = self._execute_sql(columns_t, columns_q)
+
+        wl = {}
+        if columns_whitelist:
+            wl = {c.upper(): c for c in columns_whitelist}
+            sql_columns = [c for c in sql_columns if c['name'] in wl]
+
+        cols = {wl.get(c['name'], c['name']): type_from_sql(c['type'], c['nullable']) for c in sql_columns}
+
+        return T.table(cols, name=name)
+
+    def quote_name(self, name):
+        return f'{name}'
+
+    def table_exists(self, name):
+        assert isinstance(name, Id)
+        tables = [t.lower() for t in self.list_tables()]
+        return name.lower() in tables
+
 
 class MysqlInterface(SqlInterfaceCursor):
     target = mysql
@@ -248,11 +305,8 @@ class MysqlInterface(SqlInterfaceCursor):
         names = self._execute_sql(T.list[T.string], sql_code)
         return list(map(Id, names))
 
-    def import_table_type(self, name, columns_whitelist=None):
-        assert isinstance(name, Id)
-        assert len(name.parts) == 1 # TODO !
 
-        columns_t = T.table(dict(
+    DESC_TABLE_TYPE = T.table(dict(
             name=T.string,
             type=T.string,
             nullable=T.string,
@@ -260,14 +314,19 @@ class MysqlInterface(SqlInterfaceCursor):
             default=T.string,
             extra=T.string,
         ))
+
+    def import_table_type(self, name, columns_whitelist=None):
+        assert isinstance(name, Id)
+        assert len(name.parts) == 1 # TODO !
+
         columns_q = "desc %s" % name.name
-        sql_columns = self._execute_sql(columns_t, columns_q)
+        sql_columns = self._execute_sql(self.DESC_TABLE_TYPE, columns_q)
 
         if columns_whitelist:
             wl = set(columns_whitelist)
             sql_columns = [c for c in sql_columns if c['name'] in wl]
 
-        cols = {c['name']: type_from_sql(c['type'].decode(), c['nullable']) for c in sql_columns}
+        cols = {c['name']: type_from_sql(c['type'], c['nullable']) for c in sql_columns}
 
         return T.table(cols, name=name)
 
@@ -275,6 +334,48 @@ class MysqlInterface(SqlInterfaceCursor):
         return f'`{name}`'
 
 
+class PrestoInterface(SqlInterfaceCursor):
+    target = presto
+
+    offset_before_limit = True
+
+    DESC_TABLE_TYPE = T.table(dict(
+        name=T.string,
+        type=T.string,
+        unknown_empty1=T.string,
+        unknown_empty2=T.string,
+    ))
+
+    def __init__(self, host, port, user, password, catalog, schema=None, print_sql=False):
+        self.args = dict(host=host, user=user, catalog=catalog, schema=schema)
+        super().__init__(print_sql)
+
+    def _create_connection(self):
+        import prestodb
+
+        return prestodb.dbapi.connect(**self.args)
+
+    def import_table_type(self, name, columns_whitelist=None):
+        assert isinstance(name, Id)
+        assert len(name.parts) == 1 # TODO !
+
+        columns_q = "desc %s" % name.name
+        sql_columns = self._execute_sql(self.DESC_TABLE_TYPE, columns_q)
+
+        if columns_whitelist:
+            wl = set(columns_whitelist)
+            sql_columns = [c for c in sql_columns if c['name'] in wl]
+
+        cols = {c['name']: type_from_sql(c['type'], False) for c in sql_columns}
+
+        return T.table(cols, name=name)
+
+    def quote_name(self, name):
+        return f'"{name}"'
+
+    list_tables = MysqlInterface.list_tables
+
+
 class PostgresInterface(SqlInterfaceCursor):
     target = postgres
 
@@ -371,22 +472,144 @@ class PostgresInterface(SqlInterfaceCursor):
             yield schema, table_name, T.table(cols, name=Id(schema, table_name))
 
 
+class RedshiftInterface(PostgresInterface):
+    target = redshift
+
+    id_type_decl = "INT IDENTITY(1,1)"
+
+
+class SnowflakeInterface(SqlInterface):
+    target = snowflake
+
+    id_type_decl = "number autoincrement"
+    max_rows_per_query = 16384
+
+    def __init__(self, account, user, password, warehouse, schema, database, role=None, print_sql=False):
+        import logging
+        logging.getLogger('snowflake.connector').setLevel(logging.WARNING)
+        import snowflake.connector
+
+        self._client = snowflake.connector.connect(
+            user=user,
+            password=password,
+            account=account,
+            role=role,
+            database=database,
+            warehouse=warehouse,
+            schema=schema,
+        )
+
+        self._print_sql = print_sql
+
+    def commit(self):
+        self._client.commit()
+
+    def quote_name(self, name):
+        # TODO is this right?
+        return name
+
+    def table_exists(self, name):
+        assert isinstance(name, Id)
+        tables = [t.lower() for t in self.list_tables()]
+        return name.lower() in tables
+
+    def list_tables(self):
+        sql_code = "SHOW TABLES"
+        table_type = T.table(dict(
+            created_on=T.datetime,
+            name=T.string,
+            database_name=T.string,
+            schema_name=T.string,
+            kind=T.string,
+            comment=T.string,
+            cluster_by=T.string,
+            rows=T.int,
+            bytes=T.int,
+            owner=T.string,
+            retention_time=T.string,
+            dropped_on=T.string,
+            automatic_clustering=T.string,
+            change_tracking=T.string,
+            search_optimization=T.string,
+            search_optimization_progress=T.string,
+            search_optimization_bytes=T.string,
+            # is_exteral=T.string,
+        ))
+        
+        tables = self._execute_sql(table_type, sql_code)
+        return [Id(table['name']) for table in tables]
+
+    def import_table_type(self, name, columns_whitelist=None):
+        assert isinstance(name, Id)
+        sql_code = f"DESC TABLE {quote_id(name)}"
+        table_type = T.table(dict(
+            name=T.string,
+            type=T.string,
+            kind=T.string,
+            null=T.bool,
+            default=T.string,
+            primary_key=T.bool,
+            unique_key=T.bool,
+            check=T.string,
+            expression=T.string,
+            comment=T.string,
+            policy=T.string,
+        ))
+        fields = self._execute_sql(table_type, sql_code)
+
+        cols = {
+            f['name']: type_from_sql(f['type'], f['null'] == 'Y')
+            for f in fields
+            if columns_whitelist is None or f['name'] in columns_whitelist
+        }
+
+        return T.table(cols, name=name)
+
+    def _import_result(self, sql_type, c):
+        if sql_type is T.nulltype:
+            return None
+
+        try:
+            res = c.fetchall()
+        except Exception as e:
+            msg = "Exception when trying to fetch SQL result. Got error: %s"
+            raise DatabaseQueryError(msg%(e))
+
+        return sql_result_to_python(Const(sql_type, res))
+    def _execute_sql(self, sql_type, sql_code):
+        import snowflake.connector
+        cs = self._client.cursor()
+        try:
+            res = cs.execute(sql_code)
+            return self._import_result(sql_type, res)
+        except snowflake.connector.errors.DatabaseError as e:
+            msg = "Exception when trying to execute SQL code:\n    %s\n\nGot error: %s"
+            raise DatabaseQueryError(msg%(sql_code, e))
+        finally:
+            cs.close()
+
+    def commit(self):
+        pass
+    def rollback(self):
+        pass
 
 class BigQueryInterface(SqlInterface):
     target = bigquery
 
-    PREQL_DATASET = '_preql_d9e7334a9e028e8aa38509912dfc2aac'
+    DEFAULT_PREQL_DATASET = '_preql_d9e7334a9e028e8aa38509912dfc2aac'
 
     supports_foreign_key = False
     id_type_decl = 'STRING NOT NULL'
 
-    def __init__(self, project, print_sql=False):
+    def __init__(self, project, dataset, print_sql=False):
         from google.cloud import bigquery
 
         # job_config = bigquery.job.QueryJobConfig(default_dataset=f'{project}._preql')
         # self._client = bigquery.Client(project, default_query_job_config=job_config)
         self._client = bigquery.Client(project)
-        self._default_dataset = None
+        self.project = project
+        self.dataset = dataset
+        self._active_dataset = dataset
 
         self._print_sql = print_sql
 
@@ -399,8 +622,8 @@ class BigQueryInterface(SqlInterface):
         if self._dataset_ensured:
             return
 
-        self._client.delete_dataset(self.PREQL_DATASET, delete_contents=True, not_found_ok=True)
-        self._client.create_dataset(self.PREQL_DATASET)
+        self._client.delete_dataset(self.DEFAULT_PREQL_DATASET, delete_contents=True, not_found_ok=True)
+        self._client.create_dataset(self.DEFAULT_PREQL_DATASET)
 
         self._dataset_ensured = True
 
@@ -490,8 +713,9 @@ class BigQueryInterface(SqlInterface):
     def list_datasets(self):
         return [ds.dataset_id for ds in self._client.list_datasets()]
 
-    def set_default_dataset(self, dataset):
-        self._default_dataset = dataset
+    def set_active_dataset(self, dataset):
+        self._client.create_dataset(dataset, exists_ok=True)
+        self._active_dataset = dataset
 
     def get_default_dataset(self):
         # if self._default_dataset is None:
@@ -501,7 +725,7 @@ class BigQueryInterface(SqlInterface):
         #     self._default_dataset = datasets[0]
         # return self._default_dataset
         self.ensure_dataset()
-        return self.PREQL_DATASET
+        return self.DEFAULT_PREQL_DATASET
 
     def qualified_name(self, name):
         "Ensure the name has a dataset"
@@ -509,7 +733,7 @@ class BigQueryInterface(SqlInterface):
         if len(name.parts) > 1:
             # already has dataset
             return name
-        return Id(self.get_default_dataset(), name.parts[-1])
+        return Id(self._active_dataset or self.get_default_dataset(), name.parts[-1])
 
     def rollback(self):
         # XXX No error? No warning?
@@ -726,6 +950,8 @@ def _drop_tables(state, *tables):
 
 
 _SQLITE_SCHEME = 'sqlite://'
+HELP_SNOWFLAKE_URI_FORMAT = 'snowflake://<user>:<pass>@<account>/<database>/<schema>?warehouse=<warehouse>'
+HELP_PRESTO_URI_FORMAT = 'presto://<user>@<host>/<catalog>/<schema>'
 
 def create_engine(db_uri, print_sql, auto_create):
     if db_uri.startswith(_SQLITE_SCHEME):
@@ -741,6 +967,30 @@ def create_engine(db_uri, print_sql, auto_create):
         raise NotImplementedError("Preql doesn't support multiple schemes")
     scheme ,= dsn.schemes
 
+    if scheme == 'snowflake':
+        if len(dsn.paths) == 1:
+            database, = dsn.paths
+            schema = dsn.query['schema']
+        elif len(dsn.paths) == 2:
+            database, schema = dsn.paths
+        else:
+            raise ValueError(f"Too many parts in path. Expected format: '{HELP_SNOWFLAKE_URI_FORMAT}'")
+        try:
+            warehouse = dsn.query["warehouse"]
+        except KeyError:
+            raise ValueError(f"Must provide warehouse. Expected format: '{HELP_SNOWFLAKE_URI_FORMAT}'")
+        return SnowflakeInterface(dsn.host, dsn.user, dsn.password, warehouse=warehouse, database=database, schema=schema, print_sql=print_sql)
+    elif scheme == "presto":
+        if len(dsn.paths) == 1:
+            catalog, = dsn.paths
+            schema = dsn.query.get('schema')
+        elif len(dsn.paths) == 2:
+            catalog, schema = dsn.paths
+        else:
+            raise ValueError(f"Too many parts in path. Expected format: '{HELP_PRESTO_URI_FORMAT}'")
+
+        return PrestoInterface(dsn.host, dsn.port, dsn.user, dsn.password, catalog=catalog, schema=schema, print_sql=print_sql)
+
     if len(dsn.paths) == 0:
         path = ''
     elif len(dsn.paths) == 1:
@@ -748,7 +998,7 @@ def create_engine(db_uri, print_sql, auto_create):
     else:
         raise ValueError("Bad value for uri, too many paths: %s" % db_uri)
 
-    if scheme == 'postgres':
+    if scheme == 'postgres' or scheme == 'postgresql':
         return PostgresInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
     elif scheme == 'mysql':
         return MysqlInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
@@ -757,6 +1007,10 @@ def create_engine(db_uri, print_sql, auto_create):
     elif scheme == 'duck':
         return DuckInterface(path, print_sql=print_sql)
     elif scheme == 'bigquery':
-        return BigQueryInterface(path, print_sql=print_sql)
+        return BigQueryInterface(dsn.host, path, print_sql=print_sql)
+    elif scheme == 'redshift':
+        return RedshiftInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
+    elif scheme == 'oracle':
+        return OracleInterface(dsn.host, dsn.port, path, dsn.user, dsn.password, print_sql=print_sql)
 
     raise NotImplementedError(f"Scheme {dsn.scheme} currently not supported")
